from collections import namedtuple
from datetime import date

from django.core.exceptions import ValidationError

from concentrator.change import ChangeSource, StorageHelper
from concentrator.models import DocExtraInfo
from concentrator.smev3.base.constants import CODE_ERROR, CODE_OK
from concentrator.smev3.applicationrequest.changes import (Smev3ChangeHelper,
                                                           Smev3ChangesMap,
                                                           Smev3DeclarationPrivilegeChangeHelper,
                                                           Smev3DeclarationUnitChangeHelper,
                                                           Smev3DeclarationDocsChangeHelper)
from concentrator.smev3.applicationrequest.constants import DISALLOWED_STATUSES
from kinder.core.children.models import Children, Delegate
from kinder.core.declaration.models import (Declaration, DeclarationPrivilege,
                                            DeclarationUnit)
from kinder.webservice.api.declaration import (ApiException,
                                               get_decl_by_client_id)

from .application import Application


Response = namedtuple('Response', 'order_id, status_code, comment')
MAX_AGE = 8


class _Message:
    """
    Сообщения для результата обработки заявки
    """
    SUCCESS = 'Заявление принято к рассмотрению'
    CHANGES_SUCCESS = 'Изменения приняты'
    NO_CHANGES = 'Изменений не найдено'
    DATA_ERROR = 'Ошибка в данных запроса'
    MULTIPLE = 'Заявление на ребенка с указанными параметрами существует'
    OLD = 'Превышен допустимый возраст - 8 лет'


class Step:
    """Класс шага обработки заявления."""

    number = 0

    def __init__(self, request_declaration, 
                 declaration=None, attachments=None):
        """
        :param request_declaration: XML структура заявления
        :param declaration: Сохраненное заявление в ЭДС
        :param attachments: Вложения
        """

        self.request_declaration = request_declaration
        self.declaration = declaration
        self.attachments = attachments

    def __repr__(self):
        return f'<{self.number}: {self.__class__.__name__}>'


class CheckOrderStep(Step):
    """Поиск совпадения по идентификатору заявления orderId."""

    number = 1

    def __next__(self):
        try:
            declaration = get_decl_by_client_id(
                self.request_declaration.orderId)
        except ApiException:
            return NewDeclarationStep(
                self.request_declaration, attachments=self.attachments)
        return ExistingDeclarationStep(
            self.request_declaration, declaration, 
            attachments=self.attachments)


class ExistingDeclarationStep(Step):
    """Сравнение данных заявления в ЭДС и данных в запросе."""

    number = 2

    def _unit_helper(self):
        """Возвращает хелпер проверки изменений в желаемых учреждениях."""

        helper = Smev3DeclarationUnitChangeHelper(
            DeclarationUnit, 'DeclarationUnit')
        helper.check_diff(self.request_declaration, self.declaration)
        return helper

    def _benefits_helper(self):
        """Возвращает хелпер проверки изменений льгот."""

        helper = Smev3DeclarationPrivilegeChangeHelper(
            DeclarationPrivilege, 'DeclarationPrivilege')
        helper.check_diff(self.request_declaration, self.declaration)
        return helper

    def __next__(self):
        today = date.today()
        border = date(today.year - MAX_AGE, today.month, today.day)
        birth = self.request_declaration.ChildInfo.ChildBirthDate
        if birth and birth <= border:
            return Response(
                order_id=self.request_declaration.orderId,
                status_code=CODE_ERROR, comment=_Message.OLD)

        change_map = Smev3ChangesMap()

        helpers = [
            self._unit_helper(),
            self._benefits_helper(),
        ]

        check_diff = [
            (Declaration, 'Declaration', self.declaration),
            (Children, 'Children', self.declaration.children),
        ]

        for children_delegate in (
                self.declaration.children.childrendelegate_set.all()):
            delegate = children_delegate.delegate
            check_diff.append((Delegate, 'Delegate', delegate))

        for model, model_name, instance in check_diff:
            helper = Smev3ChangeHelper(
                model, model_name, map_changes=change_map)
            helper.check_diff(self.request_declaration, instance)
            helpers.append(helper)

        if self.attachments:
            helper = Smev3DeclarationDocsChangeHelper(DocExtraInfo)
            helper.check_diff(
                self.declaration, self.attachments)
            helpers.append(helper)

        updated = StorageHelper.create_change(
            self.declaration, helpers, source=ChangeSource.UPDATE_APPLICATION)

        if not updated:
            return Response(
                order_id=self.request_declaration.orderId,
                status_code=CODE_ERROR,
                comment=_Message.NO_CHANGES)
        return Response(
            order_id=self.request_declaration.orderId,
            status_code=CODE_OK,
            comment=_Message.CHANGES_SUCCESS)


class NewDeclarationStep(Step):
    """Подача новой заявки, поиск на совпадение с существующими по обязательным параметрам."""

    number = 3

    def __next__(self):
        decls = Declaration.objects.filter(
            children__firstname=self.request_declaration.ChildInfo.ChildName,
            children__surname=self.request_declaration.ChildInfo.ChildSurname,
            children__patronymic=self.request_declaration.ChildInfo.ChildMiddleName,
            children__date_of_birth=self.request_declaration.ChildInfo.ChildBirthDate,
        )
        decls_count = decls.count()

        if decls_count == 1:
            declaration = decls.first()
            if declaration.status.code not in DISALLOWED_STATUSES:
                return ExistingDeclarationStep(
                    self.request_declaration, declaration, 
                    attachments=self.attachments)
            else:
                return NewDeclarationStep(
                    self.request_declaration, attachments=self.attachments)
        elif decls_count > 1:
            return MultipleDeclarationsStep(
                self.request_declaration, decls, 
                attachments=self.attachments)
        else:
            return RequestValidationStep(
                self.request_declaration, attachments=self.attachments)


class MultipleDeclarationsStep(Step):
    """Поиск совпадения по дополнительным параметрам."""

    number = 4

    def __next__(self):
        declarations = self.declaration.filter(
            children__dul_number=self.request_declaration.ChildInfo.ChildBirthDocRF.ChildBirthDocNumber,
            children__dul_series=self.request_declaration.ChildInfo.ChildBirthDocRF.ChildBirthDocSeries,
            children__dul_date=self.request_declaration.ChildInfo.ChildBirthDocRF.ChildBirthDocIssueDate,
        )
        if declarations.exists():
            if declarations.count() == 1:
                declaration = declarations.first()
                if declaration.status.code in DISALLOWED_STATUSES:
                    return ExistingDeclarationStep(
                        self.request_declaration, declaration,
                        attachments=self.attachments)
                else:
                    # найдено похожее заявление в архивном статусе
                    return Response(
                        order_id=self.request_declaration.orderId,
                        status_code=CODE_ERROR,
                        comment=_Message.MULTIPLE)
            else:
                # найдено заявление с совпадающим ДУЛ
                return Response(
                    order_id=self.request_declaration.orderId,
                    status_code=CODE_ERROR,
                    comment=_Message.MULTIPLE)
        return SNILSStep(self.request_declaration, 
                         attachments=self.attachments)


class SNILSStep(Step):
    """Поиск совпадения по СНИЛС."""

    number = 5

    def __next__(self):
        declarations = self.declaration.filter(
            children__snils=self.request_declaration.children_snils,
        )
        if declarations.exists():
            if declarations.count() == 1:
                return ExistingDeclarationStep(
                    self.request_declaration, declarations.first(), 
                    attachments=self.attachments)
            else:
                # найдено заявление с совпадающим СНИЛС
                return Response(
                    order_id=self.request_declaration.orderId,
                    status_code=CODE_ERROR,
                    comment=_Message.MULTIPLE)
        return RequestValidationStep(
            self.request_declaration, attachments=self.attachments)


class RequestValidationStep(Step):
    """Валидация запроса."""

    number = 6

    def __next__(self):
        try:
            today = date.today()
            border = date(today.year - MAX_AGE, today.month, today.day)
            birth = self.request_declaration.ChildInfo.ChildBirthDate
            if birth and birth <= border:
                return Response(
                    order_id=self.request_declaration.orderId,
                    status_code=CODE_ERROR, comment=_Message.OLD)
            Application(self.request_declaration, self.attachments).create()
        except ValidationError:
            return Response(
                order_id=self.request_declaration.orderId,
                status_code=CODE_ERROR, comment=_Message.DATA_ERROR)
        else:
            return Response(
                order_id=self.request_declaration.orderId,
                status_code=CODE_OK, comment=_Message.SUCCESS)
