from django.db.models import BooleanField
from django.db.models import Case
from django.db.models import Value
from django.db.models import When
from m3_ext.ui import all_components as ext
from m3_ext.ui.fields import ExtNumberField
from m3_ext.ui.misc import ExtDataStore

from kinder.core.direct.models import DRS

from .models import ApplicantAnswer
from .models import ApplicantAnswersEnum

APPLICANT_ANSWER_PARAM = 'applicant_answer'


def add_applicant_answer_column(columns, ext_result):
    """Добавляет столбец "Ответ заявителя" в реестр направлений."""

    _columns = columns.copy()
    index = [
        _columns.index(x) for x in _columns if x[0] == 'group.unit.name'
    ][0]

    _columns.insert(
        index + 1, (APPLICANT_ANSWER_PARAM, 'Ответ заявителя', 50)
    )

    return _columns


def add_applicant_answer_column_filter(win, name, ext_result):
    """
    Добавляет колоночный фильтр для столбца "Ответ заявителя"
    в реестре направлений.

    """

    applicant_answer_filter = ext.ExtDictSelectField(
        region='north',
        anchor='100%',
        ask_before_deleting=False,
        allow_blank=True,
        hide_edit_trigger=True,
        hide_clear_trigger=False,
        hide_dict_select_trigger=True,
        hide_trigger=False,
    )

    filter_store = ApplicantAnswersEnum.get_choices()
    filter_store.extend(list({2: 'Пусто'}.items()))
    applicant_answer_filter.set_store(
        ExtDataStore(filter_store)
    )

    win.add_grid_column_filter(
        APPLICANT_ANSWER_PARAM, applicant_answer_filter.render())


def add_applicant_answer_data(obj, ext_result):
    """
    Добавляет значение для атрибута столбца "Ответ заявителя" и атрибут для
    окрашивания строк в синий цвет, если "Ответ заявителя" = True и направление
    в статусе "Направлен в ДОО".

    """

    applicant_answer = ApplicantAnswer.objects.filter(direct_id=obj.id).first()

    if applicant_answer:
        # Атрибут для окрашивания строки в синий цвет
        obj.applicant_answer_highlight = (
            applicant_answer.answer and obj.status.code == DRS.REGISTER)

    answer = applicant_answer.answer if applicant_answer else ''
    # Человекочитаемое значение поля "Ответ заявителя"
    obj.applicant_answer = ApplicantAnswersEnum.values.get(answer, '')

    return obj


def applicant_answer_update_filter_list(fields_list, ext_result):
    """Обновляет список стоблцов с кастомными колоночными фильтрами."""

    return fields_list.append(APPLICANT_ANSWER_PARAM)


def apply_applicant_answer_filter(query, context, ext_result):
    """
    Фильтрация направлений через колоночный фильтр и сортировка по стоблцу
    "Ответ заявителя".

    """

    query = query.annotate(
        applicant_answer=Case(
            When(applicantanswer__isnull=True, then=Value(None)),
            When(applicantanswer__answer=True, then=Value(True)),
            When(applicantanswer__answer=False, then=Value(False)),
            output_field=BooleanField()
        )
    )

    # Фильтрация запроса
    if hasattr(context, APPLICANT_ANSWER_PARAM):
        param = getattr(context, APPLICANT_ANSWER_PARAM, None)

        if param == 'true':
            param = True
        elif param == 'false':
            param = False

        if param not in ApplicantAnswersEnum.values:
            applicant_answers = ApplicantAnswer.objects.filter(
                answer__isnull=False
            ).values_list('direct_id', flat=True)

            query = query.exclude(id__in=applicant_answers)

        else:
            applicant_answers = ApplicantAnswer.objects.filter(
                answer=param
            ).values_list('direct_id', flat=True)

            if applicant_answers:
                query = query.filter(id__in=applicant_answers)

    return query


def add_days_for_confirm_proposed_doo_field(tab, ext_result):
    """Добавление поля "Количество дней на согласование предложенного ДОО"
    на вкладку "Комплектование" в справочник "МО"

    :param tab: Вкладка, на которое будет добавлено поле
    :type tab: ObjectTab
    :param ext_result: результат выполнения предыдущего обработчика точки
        расширения (в случае, если таковых несколько)
    :type ext_result: Any
    """
    days_for_confirm_proposed_doo_field = ExtNumberField(
        anchor='100%',
        label='Количество дней на согласование предложенного ДОО',
        name='days_for_confirm_proposed_doo',
        allow_blank=False,
        allow_decimals=False,
        allow_negative=False,
        max_length=3,
        min_value=0,
        max_value=999,
    )
    tab.field__days_for_confirm_proposed_doo = (
        days_for_confirm_proposed_doo_field)

    tab.panel.items.append(tab.field__days_for_confirm_proposed_doo)
